open Parser;;

let rec parens x =
 ((char '(' >> parens >> char ')' >> parens) <|> return ()) x
;;
let rec nesting x = ((
  char '(' >>
  nesting >>= fun i ->
  char ')' >>
  nesting >>= fun j ->
  return (max (i + 1) j)
 ) <|> return 0
) x;;
let word = many1 letter;;
let separator = skip_many1 (space <|> char ',');;
let sentence =
 sep_by1 word separator >>= fun ws ->
 one_of ".?!" >>
 return ws
;;
let any_pair = any_char >>= fun c -> any_char >>= fun d -> return (c, d);;
let braced_char = between (char '{') any_char (char '}');;
let comma_chars = sep_by any_char (char ',');;

type arith =
 | Num of Strng.t
 | Add of arith * arith
 | Mul of arith * arith
;;
(* wrong: left-recursive *)
let rec e x = ( 
 (e >>= fun e1 -> char '+' >> t >>= fun e2 -> return (Add (e1, e2)))
 <|> t
) x
and t x = ( 
 (t >>= fun t1 -> char '*' >> f >>= fun t2 -> return (Mul (t1, t2)))
 <|> f
) x
and f x = ( 
 (char '(' >> e >>= fun e -> char ')' >> return e)
 <|> n
) x
and n = many1 digit >>= fun r -> return (Num r);;
(* correct: *)
let rec e x = (t >>= fun t1 -> e' t1) x
and e' term =
 (char '+' >> t >>= fun t1 -> e' t1 >>= fun t2 -> return (Add (term, t2)))
 <|> return term
and t x = (f >>= fun f1 -> t' f1) x
and t' factor =
 (char '*' >> f >>= fun f1 -> t' f1 >>= fun f2 -> return (Mul (factor, f2)))
 <|> return factor
and f x = ((char '(' >> e >>= fun r -> char ')' >> return r) <|> n) x
and n = many1 digit >>= fun r -> return (Num r);;
(* shorter: *)
let rec e x = (t >>= e') x
and e' l =
 (char '+' >> t >>= e' >>= fun r -> return (Add (l, r)))
 <|> return l 
and t x = (f >>= t') x
and t' l =
 (char '*' >> f >>= t' >>= fun r -> return (Mul (l, r)))
 <|> return l
and f x =
 ((char '(' >> e >>= fun r ->  char ')' >> return r) <|> n) x
and n = many1 digit >>= fun r -> return (Num r);;
